package mysql

import (
	"fmt"
	"sync"
	"time"

	"github.com/tidwall/gjson"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/configmdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/constantmdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/errormdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/loggermdl"
	_ "github.com/go-sql-driver/mysql"

	"github.com/gocraft/dbr"
	"github.com/tidwall/sjson"
)

// Hold a single global connection (pooling provided by sql driver)
var sqlConnections map[string]*dbr.Connection
var connectionError error
var sqlOnce sync.Once
var config tomlConfig
var defaultHost string

// MYSQLConnection -
type mysqlConnection struct {
	Server          string        `json:"server" bson:"server"`
	Username        string        `json:"username" bson:"username"`
	Password        string        `json:"password " bson:"password "`
	Protocol        string        `json:"protocol" bson:"protocol"`
	Database        string        `json:"database" bson:"database"`
	Parameters      []param       `json:"params" bson:"params"`
	MaxIdleConns    int           `json:"maxIdleConns" bson:"maxIdleConns"`
	MaxOpenConns    int           `json:"maxOpenConns" bson:"maxOpenConns"`
	ConnMaxLifetime time.Duration `json:"connMaxLifetime" bson:"connMaxLifetime"`
}
type param struct {
	ParamKey   string `json:"paramkey" bson:"paramkey"`
	ParamValue string `json:"paramvalue" bson:"paramvalue"`
}

type tomlConfig struct {
	MysqlHosts map[string]mysqlConnection
}

// Init initializes MYSQL Connections for given toml file
func Init(tomlFilepath string, defaultHostName string) (map[string]*dbr.Connection, error) {
	sqlOnce.Do(func() {
		sqlConnections = make(map[string]*dbr.Connection)
		_, err := configmdl.InitConfig(tomlFilepath, &config)
		if errormdl.CheckErr(err) != nil {
			loggermdl.LogError("Init InitConfig Err : ", err)
			connectionError = err
			return
		}
		for connectionName, connectionDetails := range config.MysqlHosts {
			paramsString := ""
			if len(connectionDetails.Parameters) > 0 {
				for paramIndex, param := range connectionDetails.Parameters {
					if paramsString == "" {
						paramsString = "?"
					}
					paramsString = paramsString + param.ParamKey + "=" + param.ParamValue
					hasNextParam := paramIndex+1 < len(connectionDetails.Parameters)
					if hasNextParam {
						paramsString = paramsString + "&"
					}
				}
			}
			//connection, err := dbr.Open("mysql", "root:root@tcp(127.0.0.1:3306)/testdb?parseTime=true", nil)
			connection, err := dbr.Open("mysql", connectionDetails.Username+":"+connectionDetails.Password+"@"+connectionDetails.Protocol+"("+connectionDetails.Server+")/"+connectionDetails.Database+paramsString, nil)
			if errormdl.CheckErr1(err) != nil {
				loggermdl.LogError("Init dbr.Open Err : ", err)
				connectionError = err
				return
			}
			if connectionDetails.MaxIdleConns == 0 {
				connectionDetails.MaxIdleConns = constantmdl.MAX_IDLE_CONNECTIONS // default is 2
			}
			if connectionDetails.MaxOpenConns == 0 {
				connectionDetails.MaxOpenConns = constantmdl.MAX_OPEN_CONNECTIONS // default there's no limit
			}
			if connectionDetails.ConnMaxLifetime == 0 {
				connectionDetails.ConnMaxLifetime = constantmdl.CONNECTION_MAX_LIFETIME
			}
			connection.SetMaxIdleConns(connectionDetails.MaxIdleConns)
			connection.SetMaxOpenConns(connectionDetails.MaxOpenConns)
			connection.SetConnMaxLifetime(connectionDetails.ConnMaxLifetime)
			sqlConnections[connectionName] = connection
		}
		defaultHost = defaultHostName
	})
	return sqlConnections, errormdl.CheckErr2(connectionError)
}

//GetMYSQLConnection -
func GetMYSQLConnection(connectionName string) (*dbr.Connection, error) {
	if errormdl.CheckBool(sqlConnections == nil) {
		loggermdl.LogError("GetMYSQLConnection Err : ", errormdl.Wrap("MYSQL_INIT_NOT_DONE"))
		return nil, errormdl.Wrap("MYSQL_INIT_NOT_DONE")
	}
	if connectionName == "" {
		if instance, keyExist := sqlConnections[defaultHost]; keyExist {
			return instance, nil
		}
	}
	if session, keyExist := sqlConnections[connectionName]; keyExist {
		return session, nil
	}
	loggermdl.LogError("GetMYSQLConnection Err : ", errormdl.Wrap("Connection not found for host: "+connectionName))
	return nil, errormdl.Wrap("Connection not found for host: " + connectionName)
}

// MysqlDAO Mysql DAO struct
type MysqlDAO struct {
	hostName string
}

// GetMysqlDAO return Mysql DAO instance
func GetMysqlDAO() *MysqlDAO {
	return &MysqlDAO{
		hostName: defaultHost,
	}
}

// GetMysqlDAOWithHost return Mysql DAO instance
func GetMysqlDAOWithHost(host string) *MysqlDAO {
	return &MysqlDAO{
		hostName: host,
	}
}

// Update Save data in Mysql db
func (md *MysqlDAO) SaveUpdateOrDeleteData(query string, data ...interface{}) error {
	connection, connectionError := GetMYSQLConnection(md.hostName)
	if errormdl.CheckErr(connectionError) != nil {
		loggermdl.LogError("SaveUpdateOrDelete GetMYSQLConnection Err : ", connectionError)
		return errormdl.CheckErr(connectionError)
	}
	stmt, stmterr := connection.Prepare(query)
	if errormdl.CheckErr1(stmterr) != nil {
		loggermdl.LogError("SaveUpdateOrDelete connection.Prepare Err : ", stmterr)
		return errormdl.CheckErr1(stmterr)
	}
	result, exeerr := stmt.Exec(data...)
	if errormdl.CheckErr2(exeerr) != nil {
		loggermdl.LogError("SaveUpdateOrDelete execution Err : ", errormdl.CheckErr2(exeerr))
		return errormdl.CheckErr2(exeerr)
	}
	_, cntErr := result.RowsAffected()
	if errormdl.CheckErr3(cntErr) != nil {
		loggermdl.LogError("SaveUpdateOrDelete count Err : ", errormdl.CheckErr3(cntErr))
		return errormdl.CheckErr3(cntErr)
	}
	// else if rowsAffected != 1 {
	// 	loggermdl.LogError("count Err : ", errormdl.CheckErr3(cntErr))
	// 	return errormdl.Wrap("mysql: expected 1 row affected, got " + strconv.Itoa(int(rowsAffected)))
	// }
	return nil
}

func (md *MysqlDAO) GetData(query string, data ...interface{}) (*gjson.Result, error) {
	connection, connectionError := GetMYSQLConnection(md.hostName)
	if errormdl.CheckErr(connectionError) != nil {
		loggermdl.LogError("GetAllData connection Err : ", connectionError)
		return nil, errormdl.CheckErr(connectionError)
	}
	rows, stmterr := connection.Query(query, data...)
	if errormdl.CheckErr1(stmterr) != nil {
		loggermdl.LogError("GetAllData query Err : ", stmterr)
		return nil, errormdl.CheckErr1(stmterr)
	}
	defer rows.Close()
	// Get column names
	columns, err := rows.Columns()
	if errormdl.CheckErr2(err) != nil {
		loggermdl.LogError("GetAllData rows.Columns() Err : ", err)
		return nil, errormdl.CheckErr2(err)
	}
	values := make([]interface{}, len(columns))
	valuePtrs := make([]interface{}, len(columns))
	tableData := "[]"
	for rows.Next() {
		for i := 0; i < len(columns); i++ {
			valuePtrs[i] = &values[i]
		}
		rows.Scan(valuePtrs...)
		data := ""
		for i, col := range columns {
			fmt.Println("col", col)
			data, err = sjson.Set(data, col, values[i])
			if errormdl.CheckErr3(err) != nil {
				loggermdl.LogError("GetAllData sjson.Set Err : ", err)
				return nil, errormdl.CheckErr3(err)
			}
		}
		tableData, err = sjson.Set(tableData, "-1", gjson.Parse(data).Value())
		if errormdl.CheckErr3(err) != nil {
			loggermdl.LogError("GetAllData sjson.Set Err : ", err)
			return nil, errormdl.CheckErr3(err)
		}
	}
	resultSet := gjson.Parse(tableData)
	return &resultSet, nil
}
